/* +------------------------------------------------------------------------+
   |                                                                        |
   |                     Dfinitions communes et prototypes                 |
   |                                                                        |
   +------------------------------------------------------------------------+ */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>


      /* +-------------------------------------------------------------+
         |  Types de donnes et constantes fonction de l'architecture  |
         +-------------------------------------------------------------+ */

#ifdef bits64

/* machine 64 bits -> codage sur 32 bits */
#define nbits  64
#define mask   0xffffffff
#define msb    0x80000000
#define shift  32
typedef unsigned long ndouble;
typedef   signed long zdouble;

#else
#ifdef bits32
#ifdef longlong

/* machine 32 bits avec donnes 64 bits -> idem */
#define nbits  64
#define mask   0xffffffffLL
#define msb    0x80000000LL
#define shift  32
typedef unsigned long long ndouble;
typedef   signed long long zdouble;

#else

/* machine 32 bits -> codage sur 16 bits */
#define nbits  32
#define mask   0xffff
#define msb    0x8000
#define shift  16
typedef unsigned long ndouble;
typedef   signed long zdouble;

#endif
#else

/* taille des mots inconnue */
#error "Unknown word size, define bits64 or bits32"

#endif
#endif


                             /* +---------------+
                                |  Dfinitions  |
                                +---------------+ */

/*
  Pour a,b entiers multiprcision on note

     high(a,b,n)  = sum( a[i]*b[j]*base^(i+j), i+j >= n)
     low(a,b,n)   = sum( a[i]*b[j]*base^(i+j), i+j <  n)
     med(a,b,n,p) = sum( a[i]*b[j]*base^(i+j), n <= i+j < p)

     divh(a,b,n) = c maximal tq high(b,c,n) <= a
*/


                        /* +-------------------------+
                           |  Fonctions utilitaires  |
                           +-------------------------+ */

/*---------- utils.c */

double  chrono();                               /* temps pass en milisecondes   */
void    dump(ndouble *a, int n);                /* affichage hexadcimal         */
ndouble *from_string(char *s, int n);           /* chane hexadcimale -> nombre */
void    random_init(ndouble *a, int n);         /* initialisation alatoire      */
int     compare(ndouble *a, ndouble *b, int n); /* compare a et b */

/*---------- memory.c */

ndouble *getmem(int n);                          /* alloue n chiffres */
void    freemem( ndouble *a);                    /* dsallocation     */
void    clearmem(ndouble *a, int n);             /* mise  zro       */
void    movemem( ndouble *a, ndouble *b, int n); /* a <- b            */


                        /* +-------------------------+
                           |  Addition/soustraction  |
                           +-------------------------+ */

/*---------- addition.c */

/* c <- a +/- b, retourne la retenue */
ndouble add(ndouble *c, ndouble *a, ndouble *b, int la, int lb);
zdouble sub(ndouble *c, ndouble *a, ndouble *b, int la, int lb);

/* a <- a +/- b, retourne la retenue */
ndouble inc(ndouble *a, ndouble *b, int la, int lb);
zdouble dec(ndouble *a, ndouble *b, int la, int lb);

/* c <- |a-b|, retourne le signe */
int asub(ndouble *c, ndouble *a, ndouble *b, int la, int lb);

/* c <- -a, retourne la retenue */
zdouble neg(ndouble *c, ndouble *a, int n);


                          /* +--------------------+
                             |  Calcul modulaire  |
                             +--------------------+ */

/*---------- modrem.c */

/* a <- b mod ... */
void mod_bp(ndouble *a, ndouble *b, int p, int n);         /* B^p - 1                        */
void mod_bq(ndouble *a, ndouble *b, int p, int q, int n);  /* ppcm(B^(p+q) - 1, .., B^p - 1) */
void mod_2p(ndouble *a, ndouble *b, int p, int t);         /* 2^p-t                          */
void msplit(ndouble *a, int p);                            /* 2^p-1, 2^p+1                   */

/* a <- b - a mod (B^p-1) */
void dec_bp(ndouble *a, ndouble *b, int p, int n);

/* combinaison de rsidus */
void mjoin( ndouble *a, int p);    /* 2^p+1, 2^p-1                       */
void mjoin1(ndouble *a, int p);    /* B^(p+1) - 1, B^p - 1               */
void mjoin2(ndouble *a, int p);    /* B^(p+2) - 1, B^(p+1) - 1, B^p - 1  */
void mjoin3(ndouble *a, int p);    /* B^(p+3) - 1, ..,          B^p - 1  */

/* dcoupage en tiers et recombinaison */
void msplit3(ndouble *a, zdouble *r, int p);              
void mcomb3( ndouble *c, ndouble *a, int p);

/* compare deux rsidus modulo B^p - 1 */
int mcompare(ndouble *a, ndouble *b, int p);

                          /* +---------------------+
                             |  Oprations en n^2  |
                             +---------------------+ */


/*---------- muln2.c */

void muln2( ndouble *c, ndouble *a, ndouble *b, int n); /* c <- ab    */
void muln2h(ndouble *c, ndouble *a, ndouble *b, int n); /* p. haute   */
void muln2l(ndouble *c, ndouble *a, ndouble *b, int n); /* p. basse   */
void muln2m(ndouble *c, ndouble *a, ndouble *b, int n); /* p. mdiane */

void sqrn2( ndouble *c, ndouble *a,             int n); /* carr      */
void sqrn2l(ndouble *c, ndouble *a,             int n); /* p. basse   */

/*---------- divn2.c */

void div2(ndouble *c, ndouble *a, int n); /* c <- a/2 */

/* q <- trunc(a/b), a <- a - q*b */
void divn2( ndouble *q, ndouble *a, ndouble *b, int n); 

/* q <- min(base^n-1, divh(a*base^(n-1), b, n-1))
   a <- a - high(b,q,n-1)/base^(n-1)
   lg(a) = p, lg(b) = n+1, lg(q) = n, p >= n+2, n >= 1, 2b >= base^(n+1) */
void divn2h(ndouble *q, ndouble *a, ndouble *b, int n, int p);

/*---------- sqrtn2.c */

/* c <- 2*trunc(a^1/2), a <- a - c^2/4, lg(a) = 2n */
void sqrtn2(ndouble *c, ndouble *a, int n);

                        /* +------------------------+
                           |  Oprations par blocs  |
                           +------------------------+ */

/*---------- blocs.c */

/* dcompose a en blocs de taille n0 chiffres + 1 chiffre de garde */
void expand(ndouble *b, ndouble *a, int la, int lb, int n0);

/* regroupe les blocs de taille n0 avec un dcalage n1 */
/* ne marche que si chaque bloc est positif            */
void compact(ndouble *b, ndouble *a,
             int la,  /* nb de blocs       */
             int n0,  /* taille des blocs  */
             int n1,  /* dcalage          */
             int lb); /* longueur rsultat */

/* addition/soustraction */
void bloc_add(ndouble *c, ndouble *a, ndouble *b,  int la, int lb, int n0);
void bloc_sub(ndouble *c, ndouble *a, ndouble *b,  int la, int lb, int n0);
void bloc_inc(ndouble *a, ndouble *b,              int la, int lb, int n0);
void bloc_dec(ndouble *a, ndouble *b,              int la, int lb, int n0);

/* multiplication par blocs en complment  2 */
void bloc_muln2(ndouble *c, ndouble *a, ndouble *b, int n, int n0);


                         /* +----------------------+
                            |  Oprations rapides  |
                            +----------------------+ */

/*---------- karatsuba.c */

/* c <- a*b, lg(a)=lg(b)=n, lg(c)=2n */
#define karamul karatsuba0
void   karamul(ndouble *c, ndouble *a, ndouble *b, int n, int n0);

void mulsz(int n, int *sa, int *sb, int *sc, int *sd);
void kara0( ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);
void kara1( ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);
int  mulchk(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);

/*---------- square.c */

/* c <- a^2, lg(a)=n, lg(c)=2n */
#define karasqr    karasqr0
void   karasqr(   ndouble *c, ndouble *a, int n, int n0);

/* c <- low(a^2), lg(a)=n, lg(c)=n+2 */
#define shortsqr_l shortsqr0
void   shortsqr_l(ndouble *c, ndouble *a, int n, int n0);

void sqrsz( int n, int *sa, int *sb, int *sc, int *sd);
void sqrlsz(int n, int *sa, int *sb, int *sc, int *sd);
void sqr0(   ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);
void sqr1(   ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);
void sqr2(   ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);
int  sqrchk( ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);
int  sqrlchk(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);


/*---------- rmp.c */

/* c <- median(a*b), lg(a)=2n-1, lg(b)=n, lg(c)=n+2 */
#define mpmul rmp0
void   mpmul(ndouble *c, ndouble *a, ndouble *b, int n, int n0);

void mpsz(int n, int *sa, int *sb, int *sc, int *sd);
void mp0(  ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);
void mp1(  ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);
int  mpchk(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);


/*---------- mulders.c */

/* c <- low(a*b), lg(a)=lg(b)=n, lg(c)=n+2 */
#define shortmul_l mulders_low
void   shortmul_l(ndouble *c, ndouble *a, ndouble *b, int n, int n0);

/* c <- high(a*b), lg(a)=lg(b)=n, lg(c)=n+1 */
#define shortmul_h mulders_high
void   shortmul_h(ndouble *c, ndouble *a, ndouble *b, int n, int n0);

/* c <- high(a*b) sans limination des retenues    */
/* lg(a)=lg(b)=lg(c)=n, 0 <= a*b -c <= (n-1)base^n */
#define highmul    raw_mulders_high
void   highmul(ndouble *c, ndouble *a, ndouble *b, int n, int n0);

void mullsz(int n, int *sa, int *sb, int *sc, int *sd);
void mulhsz(int n, int *sa, int *sb, int *sc, int *sd);
void mul0(   ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);
void mul1(   ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);
void mul2(   ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);
int  mullchk(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);
int  mulhchk(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);


/*---------- modmul.c */
#ifdef use_mul_r
void mul_r(ndouble *c, ndouble *a, ndouble *b, int n, int n0);
#else
#define mul_r karamul
#endif

/* c <- a*b mod (base^p - 1) */
void mul_bp(ndouble *c, ndouble *a, ndouble *b, int p, int la, int lb, int n0);

void mmulsz0(int n, int *sa, int *sb, int *sc, int *sd);
void mmulsz1(int n, int *sa, int *sb, int *sc, int *sd);
void mmulsz2(int n, int *sa, int *sb, int *sc, int *sd);
void mmulsz3(int n, int *sa, int *sb, int *sc, int *sd);
void mmulsz4(int n, int *sa, int *sb, int *sc, int *sd);
void mmulsz5(int n, int *sa, int *sb, int *sc, int *sd);

void mmul0(   ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);
void mmul1(   ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);
void mmul2(   ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);
void mmul3(   ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);
void mmul4(   ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);
void mmul5(   ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);

int  mmulchk0(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);
int  mmulchk1(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);
int  mmulchk2(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);
int  mmulchk3(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);


/*---------- modsqr.c */
#ifdef use_mul_r
void sqr_r(ndouble *c, ndouble *a, int n, int n0);
#else
#define sqr_r karasqr
#endif
void sqr_bp(ndouble *c, ndouble *a, int p, int n, int n0);

void msqrsz0(int n, int *sa, int *sb, int *sc, int *sd);
void msqrsz1(int n, int *sa, int *sb, int *sc, int *sd);
void msqrsz2(int n, int *sa, int *sb, int *sc, int *sd);
void msqrsz3(int n, int *sa, int *sb, int *sc, int *sd);
void msqrsz4(int n, int *sa, int *sb, int *sc, int *sd);
void msqrsz5(int n, int *sa, int *sb, int *sc, int *sd);

void msqr0(   ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);
void msqr1(   ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);
void msqr2(   ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);
void msqr3(   ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);
void msqr4(   ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);
void msqr5(   ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);

int  msqrchk0(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);
int  msqrchk1(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);
int  msqrchk2(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);
int  msqrchk3(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);

/*---------- burnikel.c */

/* c <- trunc(a*b), lg(a)=2n, lg(b)=lg(c)=n */
/* si rem=1 alors a <- a - bc               */
/* si rem=0 alors a <- a - partie(bc), le quotient peut tre surestim */
/*                terminer la division avec burnikel_correct s'il y a  */
/*                besoin de corriger c et calculer le reste exact      */
void   burnikel(int rem, ndouble *c, ndouble *a, ndouble *b, int n, int n0);
void   burnikel_correct(ndouble *c, ndouble *a, ndouble *b, int n, int n0);

/* c <- trunc(a/b), a <- a - partie(bc), le quotient retourn est exact */
void   burnikel_nr(ndouble *c, ndouble *a, ndouble *b, int n, int n0);

void burnsz(int n, int *sa, int *sb, int *sc, int *sd);
void burn0(   ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);
void burn1(   ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);
int  divnrchk(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);
int  divchk(  ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);

/*---------- fastrem.c */

/* c <- a - b*q mod ppbm(B^(lc/4) - 1, .., B^(lc/4+3) - 1) */
int remdiv(ndouble *c, ndouble *a, ndouble *b, ndouble *q,
	   int     lc, int     la, int     lb, int     lq, int n0);

/* c <- a - b^2 mod ppbm(B^(lc/4) - 1, .., B^(lc/4+3) - 1) */
int remsqrt(ndouble *c, ndouble *a, ndouble *b,
	    int     lc, int     la, int     lb, int n0);

/*---------- kmrdiv.c */

/* division modulaire */
void rm_divh1(ndouble *c, ndouble *a, ndouble *b, int n, int p, int n0);

void kmdsz0( int n, int *sa, int *sb, int *sc, int *sd);
void kmdsz1( int n, int *sa, int *sb, int *sc, int *sd);
void rdivsz2(int n, int *sa, int *sb, int *sc, int *sd);

void kmd0(    ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);
void kmd1(    ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);
void kmd2(    ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);
void rdiv0(   ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);
void rdiv1(   ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);
void rdiv2(   ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);
void rdiv3(   ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);

int  divhchk( ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);
int  invhchk( ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);
int  mdivhchk(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);

/*---------- zimmermann.c */
void sqrtsz(int n, int *sa, int *sb, int *sc, int *sd);
void zimm0(    ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);
void zimm1(    ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);
int  sqrtnrchk(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);
int  sqrtchk(  ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);

/*---------- fastsqrt.c */
void sqrt0(   ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);
void sqrt1(   ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);
int  msqrtchk(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0);
