/* +------------------------------------------------------------------------+
   |                                                                        |
   |                            Division de Knuth                           |
   |                                                                        |
   +------------------------------------------------------------------------+ */

#include "defs.h"

/* c <- a/2 */
void div2(ndouble *c, ndouble *a, int n) {
  int i;
  ndouble r;

  for (i=n-1,r=0; i>=0; i--) {
    r += a[i];
    c[i] = (r/2) & mask;
    if (r%2) {r = (ndouble)mask + 1;} else {r = 0;}
  }
}

/* q <- trunc(a/b), a <- a - q*b, lg(a) = 2n, lg(b) = lg(q) = n */
/* le chiffre de tte de b doit avoir son bit de poids fort  1 */
void divn2(ndouble *q, ndouble *a, ndouble *b, int n) {
  int   i,j;
  ndouble r,s,x;
  zdouble  z;

  /* division 2 chiffres par 1 chiffre */
  if (n <= 1) {
    if (a[1] >= b[0]) {
      q[0] = mask;
      r = a[0]+b[0] + ((a[1]-b[0]) >> shift);
      a[0] = r & mask;
      a[1] = r >> shift;
    }
    else {
      r = (a[1] << shift) + a[0];
      q[0] = r / b[0];
      a[1] = 0;
      a[0] = r % b[0];
    }
  }

  /* division  plusieurs chiffres */
  else {
    for (i=n-1; i>=0; i--) {

      /* divise les trois chiffres de tte de a par les deux chiffres de b */
      if (a[n+i] >= b[n-1]) x = mask;
      else {
        r = (a[n+i] << shift) + a[n+i-1];
        x = r / b[n-1];
        r = ((r%b[n-1]) << shift) + a[n+i-2];
	s = x * b[n-2];
	while (r < s) {x--; s -= r; r = (b[n-1] << shift) + b[n-2];}
      }
      
      /* a <- a - x*base^i */
      for (j=0,r=0; j < n; j++) {
        r += x*b[j];
        z = a[i+j] - (r&mask);
        a[i+j] = z & mask;
        r >>= shift; if (z < 0) r++;
      }
      if (r != a[n+i]) {inc(a+i,b,n,n); x--;} /* si ngatif ajoute b*base^i */
      a[n+i] = 0;
      
      q[i] = x;
    }
  }
}



                      /* +------------------+
                         |  Division haute  |
                         +------------------+ */

/* Entres
     1 <= n
     a de longueur p >= n+2
     b de longueur n+1 avec 2b >= base^(n+1)

   Sorties
     q <- min(base^n - 1, divh(a*base^(n-1), b, n-1)) de longueur n
     a <- a - high(b,q,n-1)/base^(n-1)
   
   si on a a < b*base en entre, alors a < base^3 en sortie
 */
void divn2h(ndouble *q, ndouble *a, ndouble *b, int n, int p) {
  int i,j;
  ndouble t,s,x;
  zdouble z;

  /* troncature si a >= b*base */
  while ((p > n+2) && (a[p-1] == 0)) p--;
  if ((p > n+2) || (compare(a+1,b,n+1) >= 0)) {
    for (i=0; i < n; q[i++] = mask);   /* q <- base^n - 1 */
    dec(a+1, b, p-1, n+1);             /* a <- a - b*base */
    for (i=t=0; i < n; t += b[i++]);
    inc(a,   &t,  p,   1);             /*        + b[0] + .. + b[n-1] */
    inc(a+1, b+n, p-1, 1);             /*        + b[n]*base          */
    return;
  }

  /* sinon effectue la division */
  else {

    clearmem(q,n);
    for (i=n-1; i >= 0; i--, b++) {

      /* divise les trois chiffres de tte de a par les deux chiffres de b */
      if ((p > i+3) || (a[i+2] >= b[i+1])) x = mask;
      else {
	t = (a[i+2] << shift) + a[i+1];
	x = t / b[i+1];
	t = ((t%b[i+1]) << shift) + a[i];
	s = x * b[i];
	while (t < s) {x--; s -= t; t = (b[i+1] << shift) + b[i];}
      }
      
      /* a <- a - high(x*base^i) */
      for (j=0, t=0; j <= i+1; j++) {
	t += x*b[j];
	z  = a[j] - (t&mask);
	a[j] = z & mask;
	t >>= shift; if (z < 0) t++;
      }
      if (p > i+2) {z = a[i+2] - t; a[i+2] = z & mask; z >>= shift;}
      else z = -t;

      /* si < 0, corrige */
      if (z) {inc(a,b,p,i+2); x--;}

      /* quotient dfinitif */
      q[i] = x;

      /* recalcule la longueur de a */
      while ((p > i+2) && (a[p-1] == 0)) p--;

    }
  }
}
