/* +------------------------------------------------------------------------+
   |                                                                        |
   |      Reste de division et racine carre par multiplication modulaire   |
   |                                                                        |
   +------------------------------------------------------------------------+ */

#include "defs.h"

static ndouble un = 1;


                          /* +---------------------+
                             |  reste de division  |
                             +---------------------+ */

/* Entres
    1 <= la,lb,lc,lq <= base
    a,b,q de longueur la,lb,lq
    
  Sortie
    c <- a - b*q mod ppcm(B^(lc/4) - 1, .., B^(lc/4+3) - 1) tronqu  lc chiffres
    retourne 1 si le rsultat dborde, 0 s'il tient

*/
int remdiv(ndouble *c, ndouble *a, ndouble *b, ndouble *q,
            int     lc, int     la, int     lb, int     lq, int n0) {
  int p,i,r;
  ndouble *buff;

  /* on veut p = 3 mod 6, p >= 9, 4*p > lc */
  p = (lc+4)/4; if (p < 9) p = 9; if ((p+3)%6) p += 6 - ((p+3)%6);
  buff = getmem(4*p+6+shift);

  /* calcule les rsidus de a - bq mod B^(p+3) - 1, .., B^p - 1 */
  mul_bp(buff,       b, q, p+3, lb, lq, n0); dec_bp(buff,       a, p+3, la);
  mul_bp(buff+p+3,   b, q, p+2, lb, lq, n0); dec_bp(buff+p+3,   a, p+2, la);
  mul_bp(buff+2*p+5, b, q, p+1, lb, lq, n0); dec_bp(buff+2*p+5, a, p+1, la);
  mul_bp(buff+3*p+6, b, q, p,   lb, lq, n0); dec_bp(buff+3*p+6, a, p,   la);

  /* rsultat tronqu  lc chiffres */
  mjoin3(buff, p); movemem(c, buff, lc);
  for (r=0, i=lc; (i < 4*p+6) && (!r); r = (buff[i++] != 0));

  /* si le rsidu dborde, soustrait ppcm(B^p - 1, .., B^(p+3) - 1) */
  /* on calcule ce ppcm  coup de mjoin3 ...  amliorer            */
  if (r) {
    for (i=0; i < 4*p+6; buff[i++] = mask);
    buff[0]--; buff[p+3]--; buff[2*p+5]--; buff[3*p+6]--;
    mjoin3(buff,p); inc(buff, &un, lc, 1);
    dec(c, buff, lc, lc);
  }

  freemem(buff);
  return(r);

}



                       /* +--------------------------+
                          |  reste de racine carre  |
                          +--------------------------+ */

/* Entres
    1 <= la,lb,lc <= base
    a,b de longueur la,lb
    
  Sortie
    c <- a - b^2 mod ppcm(B^(lc/4) - 1, .., B^(lc/4+3) - 1) tronqu  lc chiffres
    retourne 1 si le rsultat dborde, 0 s'il tient

*/
int remsqrt(ndouble *c, ndouble *a, ndouble *b,
            int     lc, int     la, int     lb, int n0) {
  int p,i,r;
  ndouble *buff;

  /* on veut p = 3 mod 6, p >= 9, 4*p > lc */
  p = (lc+4)/4; if (p < 9) p = 9; if ((p+3)%6) p += 6 - ((p+3)%6);
  buff = getmem(4*p+6+shift);

  /* calcule les rsidus de a - b^2 mod B^(p+3) - 1, .., B^p - 1 */
  sqr_bp(buff,       b, p+3, lb, n0); dec_bp(buff,       a, p+3, la);
  sqr_bp(buff+p+3,   b, p+2, lb, n0); dec_bp(buff+p+3,   a, p+2, la);
  sqr_bp(buff+2*p+5, b, p+1, lb, n0); dec_bp(buff+2*p+5, a, p+1, la);
  sqr_bp(buff+3*p+6, b, p,   lb, n0); dec_bp(buff+3*p+6, a, p,   la);

  /* rsultat tronqu  lc chiffres */
  mjoin3(buff, p); movemem(c, buff, lc);
  for (r=0, i=lc; (i < 4*p+6) && (!r); r = (buff[i++] != 0));

  /* si le rsidu dborde, soustrait ppcm(B^p - 1, .., B^(p+3) - 1) */
  /* on calcule ce ppcm  coup de mjoin3 ...  amliorer            */
  if (r) {
    for (i=0; i < 4*p+6; buff[i++] = mask);
    buff[0]--; buff[p+3]--; buff[2*p+5]--; buff[3*p+6]--;
    mjoin3(buff,p); inc(buff, &un, lc, 1);
    dec(c, buff, lc, lc);
  }

  freemem(buff);
  return(r);

}
