/* +------------------------------------------------------------------------+
   |                                                                        |
   |                             Multiplication modulaire                   |
   |                                                                        |
   +------------------------------------------------------------------------+ */

#include "defs.h"


             /* +------------------------------------+
                |  Multiplication modulo base^p - 1  |
                +------------------------------------+ */

/* Entres
    1 <= la,lb,p <= base
    a,b de longueurs la, lb

   Sortie
    c <- a*b mod (base^p - 1) de longueur p + shift

   Remarque
    le rsultat tient sur p chiffres, mais n'est pas normalis
*/

void mul_bp(ndouble *c, ndouble *a, ndouble *b, int p, int la, int lb, int n0) {
  ndouble *buff = getmem(2*p+2 + 2*(7*p/6+5)), *x, *y, *z;
  zdouble ra[4], rb[4];
  int i,k,l,m,q;

  /* rduit les arguments modulo base^p - 1 */
  mod_bp(buff+2*p+2, a, p, la); a = buff + 2*p+2;   a[p] = 0;
  mod_bp(a+7*p/6+5,  b, p, lb); b = a    + 7*p/6+5; b[p] = 0;

  /* dcoupe en deux tant que le nb de bits est pair et que les oprandes
     sont assez grands. Accumule les produits des rsidus modulo 2^q+1 dans c */
  for (q = p*shift; ((q/shift) > n0) && ((q & 1) == 0); ) {

    /* dcoupage */
    q >>= 1; l = (q+shift-1)/shift;
    msplit(a,q);
    msplit(b,q); 

    /* multiplication des rsidus modulo 2^q + 1 : normalement ils tiennent
       sur l chiffres sauf si q est un multiple de shift et l'un des oprandes
       est gal  2^q mod (2^q + 1). Dans ce cas, on effectue la multiplication
        la main. */
    if (((q%shift) == 0) && ((a[2*l]) || (b[2*l]))) {
      clearmem(buff,l);
      movemem(buff+l, (a[2*l]) ? b+l : a+l, l+1);
      mod_2p(c, buff, q, -1);
    }

    /* q non divisible par 3 ou trop petit -> produit ordinaire et rduction */
    else if ((q%3) || (q < 3*shift)) {
      mul_r(buff, a+l, b+l, l, n0); buff[2*l] = 0;
      mod_2p(c, buff, q, -1);
    }

    /* q = 3k -> calcul modulo (2^k + 1) et (4^k - 2^k + 1) puis combinaison */
    else {
      k = q/3; m = (k+shift-1)/shift;

      msplit3(a+l, ra, k);
      msplit3(b+l, rb, k);

      for (i=0, x=buff, y=a+l, z=b+l; i < 4; i++, x += 2*m+1, y += m, z += m) {
	mul_r(x, y, z, m, n0);
	x[2*m] = (ra[i]*rb[i]) & mask;
	if (ra[i] < 0) dec(x+m, z, m+1, m);
	if (ra[i] > 0) inc(x+m, z, m+1, m);
	if (rb[i] < 0) dec(x+m, y, m+1, m);
	if (rb[i] > 0) inc(x+m, y, m+1, m);
      }
      mcomb3(c, buff, k);
    }

    a[l] = b[l] = 0;
    c += q/shift + 1;
  }

  /* effectue le dernier produit modulo 2^q - 1. On pourrait gagner un
     peu lorsque q = 3k en calculant les produits modulo 2^k - 1 et
     4^k + 2^k + 1, mais ici soit on a divis p par au moins 16, soit
     q est trop petit, dans les deux cas le gain est minime et ne vaut
     pas la complication du code. */
  l  = (q+shift-1)/shift;
  mul_r(buff, a, b, l, n0); buff[2*l] = 0;
  mod_2p(c, buff, q, 1);

  /* recompose le produit */
  for (; q < p*shift; q <<= 1) {c -= q/shift + 1; mjoin(c,q);}

  freemem(buff);

}

           /* +---------------------------------------------------+
              |  multiplication mod ppcm(base^p-1, base^(p+1)-1)  |
              +---------------------------------------------------+ */

/* Entres
    1 <= la,lb,p <= base
    a,b de longueurs la, lb

   Sortie
    c <- a*b mod ppcm(base^p - 1, base^(p+1) - 1) de longueur 2p+1 + shift

   Remarques
    ppcm = (base^p - 1)*(base^(p+1) - 1)/(base - 1)
    le rsultat est normalis

*/
void mul_p1(ndouble *c, ndouble *a, ndouble *b, int p, int la, int lb, int n0) {

  mul_bp(c,     a, b, p+1, la, lb, n0);
  mul_bp(c+p+1, a, b, p,   la, lb, n0);
  mjoin1(c,p);

}

   /* +-----------------------------------------------------------------+
      |  multiplication mod ppcm(base^p-1, base^(p+1)-1, base^(p+2)-1)  |
      +-----------------------------------------------------------------+ */

/* Entres
    4 <= p <= base, p pair
    1 <= la,lb <= base
    a,b de longueurs la, lb

   Sortie
    c <- a*b mod ppcm(base^p - 1, base^(p+1) - 1, base^(p+2) - 1)
                de longueur 3p+3 + shift

  Remarques
    ppcm = (B^p - 1)*(B^(p+1) - 1)*(B^(p+2) - 1)/(B - 1)/(B^2 - 1)
    le rsultat est normalis

*/
void mul_p2(ndouble *c, ndouble *a, ndouble *b, int p, int la, int lb, int n0) {

  mul_bp(c,       a, b, p+2, la, lb, n0);
  mul_bp(c+p+2,   a, b, p+1, la, lb, n0);
  mul_bp(c+2*p+3, a, b, p,   la, lb, n0);
  mjoin2(c,p);

}

   /* +-------------------------------------------------------+
      |  multiplication mod ppcm(base^p-1, .., base^(p+3)-1)  |
      +-------------------------------------------------------+ */

/* Entres
    9 <= p <= base, p impair multiple de 3
    1 <= la,lb <= base
    a,b de longueurs la, lb

   Sortie
    c <- a*b mod ppcm(base^p - 1, .., base^(p+3) - 1)
                de longueur 4p+6 + shift

  Remarques
    ppcm = produit/(B - 1).(B^2 - 1).(B^3 - 1)
    le rsultat est normalis
*/
void mul_p3(ndouble *c, ndouble *a, ndouble *b, int p, int la, int lb, int n0) {

  mul_bp(c,       a, b, p+3, la, lb, n0);
  mul_bp(c+p+3,   a, b, p+2, la, lb, n0);
  mul_bp(c+2*p+5, a, b, p+1, la, lb, n0);
  mul_bp(c+3*p+6, a, b, p,   la, lb, n0);
  mjoin3(c,p);

}

                 /* +----------------------------+
                    |  Multiplication rcursive  |
                    +----------------------------+ */

/* Entres
    1 <= n <= base
    a,b de longueur n

  Sortie
    c = a*b de longueur 2*n
*/

#ifdef use_mul_r
void mul_r(ndouble *c, ndouble *a, ndouble *b, int n, int n0) {

  /* petits oprandes -> multiplication usuelle */
  if (n <= 6*n0) karamul(c,a,b,n,n0);

  /* sinon, calcul modulaire en 4 points */
  else {
    int p;
    ndouble *buff;

    /* on veut p = 3 mod 6, p >= 9, 4*p >= 2*n */
    p = (2*n+3)/4; if (p < 9) p = 9; if ((p+3)%6) p += 6 - ((p+3)%6);
    buff = getmem(4*p+6+shift);

    /* multiplication modulaire */
    mul_p3(buff, a, b, p, n, n, n0);    
    movemem(c, buff, 2*n);

    freemem(buff);
  }
    
}
#endif

                           /* +---------+
                              |  tests  |
                              +---------+ */

/* multiplication modulo B^p - 1 */
static void size0(int n, int *sa, int *sb, int *sc, int *sd, int *p) {
  int l;

  *p = n; l = *p;

  *sa = (5*n + 1)/2;
  *sb = (5*n + 1)/2;
  *sc = -(l + shift);
  *sd = 0;
}

void mmulsz0(int n, int *sa, int *sb, int *sc, int *sd) {
  int p;
  size0(n, sa, sb, sc, sd, &p);
}

void mmul0(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {
  int sa, sb, sc, sd, p;

  size0(n, &sa, &sb, &sc, &sd, &p);
  mul_bp(c, a, b, p, sa, sb, n0);
}

int mmulchk0(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {
  ndouble *e, *f;
  int sa, sb, sc, sd, p, res;

  size0(n, &sa, &sb, &sc, &sd, &p);
  e = getmem(2*sa);
  f = getmem(-sc);

  karamul(e, a, b, sa, n0);
  mod_bp(f, e, p, 2*sa);  res = mcompare(c, f, p);

  freemem(e);
  freemem(f);
  return(res);
}

/* multiplication modulo B^p - 1, B^(p+1) - 1 */
static void size1(int n, int *sa, int *sb, int *sc, int *sd, int *p) {
  int l;

  *p = (n+1)/2; l = 2*(*p)+1;

  *sa = (5*n + 1)/2;
  *sb = (5*n + 1)/2;
  *sc = -(l + shift);
  *sd = 0;
}

void mmulsz1(int n, int *sa, int *sb, int *sc, int *sd) {
  int p;
  size1(n, sa, sb, sc, sd, &p);
}

void mmul1(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {
  int sa, sb, sc, sd, p;

  size1(n, &sa, &sb, &sc, &sd, &p);
  mul_p1(c, a, b, p, sa, sb, n0);
}

int mmulchk1(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {
  ndouble *e, *f, *g;
  int sa, sb, sc, sd, p, res;

  size1(n, &sa, &sb, &sc, &sd, &p);
  e = getmem(2*sa);
  f = getmem(p + 1 + shift);
  g = getmem(p + 1 + shift);

  karamul(e, a, b, sa, n0);
  mod_bp(f, e, p,   2*sa); mod_bp(g, c, p,   2*p+1); res  = mcompare(f, g, p);
  mod_bp(f, e, p+1, 2*sa); mod_bp(g, c, p+1, 2*p+1); res |= mcompare(f, g, p+1);

  freemem(e); freemem(f); freemem(g);

  return(res);
}

/* multiplication modulo B^p - 1, B^(p+1) - 1, B^(p+2) - 1 */
static void size2(int n, int *sa, int *sb, int *sc, int *sd, int *p) {
  int l;

  *p = (n+2)/3; *p += (*p%2); if (*p < 4) *p=4; l = 3*(*p)+3;

  *sa = (5*n + 1)/2;
  *sb = (5*n + 1)/2;
  *sc = -(l + shift);
  *sd = 0;
}

void mmulsz2(int n, int *sa, int *sb, int *sc, int *sd) {
  int p;
  size2(n, sa, sb, sc, sd, &p);
}

void mmul2(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {
  int sa, sb, sc, sd, p;

  size2(n, &sa, &sb, &sc, &sd, &p);
  mul_p2(c, a, b, p, sa, sb, n0);
}

int mmulchk2(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {
  ndouble *e, *f, *g;
  int sa, sb, sc, sd, p, res;

  size2(n, &sa, &sb, &sc, &sd, &p);
  e = getmem(2*sa);
  f = getmem(p + 2 + shift);
  g = getmem(p + 2 + shift);

  karamul(e, a, b, sa, n0);
  mod_bp(f, e, p,   2*sa); mod_bp(g, c, p,   3*p+3); res  = mcompare(f, g, p);
  mod_bp(f, e, p+1, 2*sa); mod_bp(g, c, p+1, 3*p+3); res |= mcompare(f, g, p+1);
  mod_bp(f, e, p+2, 2*sa); mod_bp(g, c, p+2, 3*p+3); res |= mcompare(f, g, p+2);

  freemem(e); freemem(f); freemem(g);

  return(res);
}

/* multiplication modulo B^p - 1, B^(p+1) - 1, B^(p+2) - 1, B^(p+3) - 1 */
static void size3(int n, int *sa, int *sb, int *sc, int *sd, int *p) {
  int l;

  *p = (n+3)/4;  while (((*p%6) != 3) || (*p < 9)) (*p)++; l = 4*(*p)+6;

  *sa = (5*n + 1)/2;
  *sb = (5*n + 1)/2;
  *sc = -(l + shift);
  *sd = 0;
}

void mmulsz3(int n, int *sa, int *sb, int *sc, int *sd) {
  int p;
  size3(n, sa, sb, sc, sd, &p);
}

void mmul3(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {
  int sa, sb, sc, sd, p;

  size3(n, &sa, &sb, &sc, &sd, &p);
  mul_p3(c, a, b, p, sa, sb, n0);
}

int mmulchk3(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {
  ndouble *e, *f, *g;
  int sa, sb, sc, sd, p, res;

  size3(n, &sa, &sb, &sc, &sd, &p);
  e = getmem(2*sa);
  f = getmem(p + 3 + shift);
  g = getmem(p + 3 + shift);

  karamul(e, a, b, sa, n0);
  mod_bp(f, e, p,   2*sa); mod_bp(g, c, p,   4*p+6); res  = mcompare(f, g, p);
  mod_bp(f, e, p+1, 2*sa); mod_bp(g, c, p+1, 4*p+6); res |= mcompare(f, g, p+1);
  mod_bp(f, e, p+2, 2*sa); mod_bp(g, c, p+2, 4*p+6); res |= mcompare(f, g, p+2);
  mod_bp(f, e, p+3, 2*sa); mod_bp(g, c, p+3, 4*p+6); res |= mcompare(f, g, p+3);

  freemem(e); freemem(f); freemem(g);

  return(res);
}


/* multiplication complte en 3 points */
static void size4(int n, int *sa, int *sb, int *sc, int *sd, int *p) {
  int l;

  *p = (2*n+2)/3; *p += (*p%2); if (*p < 4) *p=4; l = 3*(*p)+3;

  *sa = n;
  *sb = n;
  *sc = -(l + shift);
  *sd = 0;
}

void mmulsz4(int n, int *sa, int *sb, int *sc, int *sd) {
  int p;
  size4(n, sa, sb, sc, sd, &p);
}

void mmul4(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {
  int sa, sb, sc, sd, p;

  size4(n, &sa, &sb, &sc, &sd, &p);
  mul_p2(c, a, b, p, sa, sb, n0);
}

/* multiplication complte en 4 points */
static void size5(int n, int *sa, int *sb, int *sc, int *sd, int *p) {
  int l;

  *p = (2*n+3)/4; while (((*p%6) != 3) || (*p < 9)) (*p)++; l = 4*(*p)+6;

  *sa = n;
  *sb = n;
  *sc = -(l + shift);
  *sd = 0;
}

void mmulsz5(int n, int *sa, int *sb, int *sc, int *sd) {
  int p;
  size5(n, sa, sb, sc, sd, &p);
}

void mmul5(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {
  int sa, sb, sc, sd, p;

  size5(n, &sa, &sb, &sc, &sd, &p);
  mul_p3(c, a, b, p, sa, sb, n0);
}

