/* +------------------------------------------------------------------------+
   |                                                                        |
   |                               Calcul modulaire                         |
   |                                                                        |
   +------------------------------------------------------------------------+ */

#include "defs.h"

static ndouble un = 1;

                /* +-------------------------------+
                   |  Rduction modulo base^p - 1  |
                   +-------------------------------+ */

/* Entres
     1 <= n,p <= base
     b de longueur n

   Sortie
     a <- b mod (base^p - 1)

   Remarque
     a ne doit pas recouvrir b
     le rsultat n'est pas normalis
*/
void mod_bp(ndouble *a, ndouble *b, int p, int n) {
  int i;
  ndouble r;

  /* copie le dernier bloc */
  i = (n%p) ? n-(n%p) : n-p;
  b += i; n -= i; i /= p;
  movemem(a,b,n); clearmem(a+n,p-n);

  /* additionne les blocs prcdents */
  for (r=0; i; i--) {b -= p; r += inc(a,b,p,p);}

  /* rincopore la retenue */
  while (r) r = inc(a,&r,p,1);

}

          /* +-------------------------------------------+
             |  Dcomposition modulo 2^p - 1 et 2^p + 1  |
             +-------------------------------------------+ */

/* Entres
     1 <= p <= base
     a de longueur ceil(p/shift) + ceil((p+1)/shift)

  Sorties
     a <- [a mod (2^p - 1), a mod (2^p + 1)]

  Remarques
   Si p n'est est pas multiple de shift :
     les rsidus sont rduits  ceil(p/shift) chiffres, ils ne sont
     pas normaliss.
   Si p est multiple de shift :
     le rsidu modulo 2^p - 1 n'est pas normalis, le rsidu modulo
     2^p + 1 l'est.
*/
void msplit(ndouble *a, int p) {

  /* si p bits = nb entier de chiffres -> addition/soustraction */
  if ((p%shift) == 0) {
    int i, l=p/shift;
    ndouble *b = a+l, r;
    zdouble z;

    /* a <- a+b, b <- a - b */
    for (i=r=z=0; i < l; i++) {
      r += a[i] + b[i]; z += a[i] - b[i];
      a[i] = r & mask;  b[i] = z & mask;
      r >>= shift;      z >>= shift;
    }

    /* propage les retenues */
    for (r += b[l]; r; r = inc(a, &r, l, 1));
    r = b[l] - z;   b[l] = inc(b, &r, l, 1);
  }

  /* sinon, addition/soustraction avec dcalage */
  else {
    int i, l = p/shift, s = shift - (p%shift);
    ndouble *b = a+l+1, r, x,y;
    zdouble z;

    /* a <- a + (b << s), b <- a - (b << s) */
    for (i=r=z=0, x = b[-1] >> (shift - s); i < l; i++) {
      x += b[i] << s;
      r += a[i] + (x & mask); z += a[i] - (x & mask);
      a[i] = r & mask;        b[i] = z & mask;
      r >>= shift;            z >>= shift;
      x >>= shift;
    }
    /* derniers chiffres */
    y  = a[l] & ((1 << (shift-s)) - 1);
    x += b[l] << s;
    r += y + x;      z += y - x;
    a[l] = r & mask; b[l] = z & mask;
    r >>= shift;     z >>= shift;

    /* propage les retenues */
    for (r <<= s; r; r = inc(a, &r, l+1, 1) << s);
    r = (-z) << s;   r = inc(b, &r, l+1, 1) << (s-1);
    if (r) {b[l] |= msb; dec(b, &r, l+1, 1);}
  }
    
}

               /* +--------------------------------+
                  |  Dcomposition modulo 2^p - t  |
                  +--------------------------------+ */

/* Entres
     1 <= p <= base
     t = +/- 1
     b de longueur ceil(p/shift) + ceil((p+1)/shift)

  Sorties
     a <- b mod (2^p - t)

  Remarque
     code adapt de msplit, mmes remarques
*/
void mod_2p(ndouble *a, ndouble *b, int p, int t) {

  /* si p bits = nb entier de chiffres -> addition/soustraction */
  if ((p%shift) == 0) {
    int i, l=p/shift;
    ndouble *c = b+l, r;
    zdouble z;

    if (t == 1) { /* a <- b + c */
      for (i=r=0; i < l; i++) {r += b[i] + c[i]; a[i] = r & mask; r >>= shift;}
      for (r += c[l]; r; r = inc(a, &r, l, 1));
    }
    else {        /* a <- b - c */
      for (i=z=0; i < l; i++) {z += b[i] - c[i]; a[i] = z & mask; z >>= shift;}
      r = c[l] - z; a[l] = inc(a, &r, l, 1);
    }
  }

  /* sinon, addition/soustraction avec dcalage */
  else {
    int i, l = p/shift, s = shift - (p%shift);
    ndouble *c = b+l+1, r, x,y;
    zdouble z;

    if (t == 1) { /* a <- b + (c << s) */

      for (i=r=0, x = c[-1] >> (shift - s); i < l; i++) {
        x += c[i] << s;
        r += b[i] + (x & mask);
        a[i] = r & mask;
        r >>= shift;
        x >>= shift;
      }

      y  = b[l] & ((1 << (shift-s)) - 1);
      x += c[l] << s;
      r += y + x;
      a[l] = r & mask;
      r >>= shift;

      for (r <<= s; r; r = inc(a, &r, l+1, 1) << s);
    }

    else {        /* a <- b - (c << s) */
      for (i=z=0, x = c[-1] >> (shift - s); i < l; i++) {
        x += c[i] << s;
        z += b[i] - (x & mask);
        a[i] = z & mask;
        z >>= shift;
        x >>= shift;
      }

      y  = b[l] & ((1 << (shift-s)) - 1);
      x += c[l] << s;
      z += y - x;
      a[l] = z & mask;
      z >>= shift;

      r = (-z) << s;   r = inc(a, &r, l+1, 1) << (s-1);
      if (r) {a[l] |= msb; dec(a, &r, l+1, 1);}
    }
  }
}

             /* +----------------------------------------------+
                |  dcomposition mod 2^p + 1 et 4^p - 2^p + 1  |
                +----------------------------------------------+ */

/* Entres
    p >= shift
    a = [a0,a1,a2] de longueur ceil(3p/shift) rsidu mod 8^p + 1

  Sorties
   a <- [u,v,w,x] de longueur 4*l, l = ceil(p/shift)
   r <- [ru,rv,rw,rx] retenues valant 0,1,-1
   avec u + ru*base^l = a0 - a1 + a2
        v + rx*base^l = a1 + a2
        w + rv*base^l = a0 + a1
        x + rw*base^l = a0 - a2
*/
void msplit3(ndouble *a, zdouble *r, int p) {
  int l = (p+shift-1)/shift;
  int i;

  /* p bits = nb non entier de chiffres : normalise et spare a0,a1,a2 */
  if (p%shift) {
    int m = (3*p+shift-1)/shift, s;
    ndouble *x, *y;
    zdouble z;

    /* normalise a modulo 8^p + 1 */
    s = (3*p)%shift;
    z = a[m-1] >> s; a[m-1] &= ((1<<s)-1);
    z = dec(a, (ndouble *)(&z),m,1);
    if (z) {inc(a, &un, m, 1); z = a[m-1] + (1<<s); a[m-1] = z & mask;}

    /* dcale a2 */
    x = a + ((2*p)/shift); y = a + 2*l;
    z = (((2*p)/shift) + l < m) ? x[l] : 0;
    s = (2*p)%shift;
    for (i=l-1; i >= 0; i--) {
      z = (z << shift) + x[i];
      y[i] = (z >> s) & mask;
    }
    x[0] &= (1<<s) - 1;

    /* dcale a1 */
    x = a + (p/shift); y = a + l;
    z = ((p/shift) + l <= (2*p)/shift) ? x[l] : 0;
    s = p%shift;
    for (i = l-1; i >= 0; i--) {
      z = (z << shift) + x[i];
      y[i] = (z >> s) & mask;
    }

    /* supprime les bits en trop dans a0 */
    a[l-1] &= (1<<s) - 1;
  }

  /* calcule u,v,w,x et les retenues */
  {
    ndouble *u = a, *v = u+l, *w = v+l, *x = w+l;
    for (i = r[0] = r[1] = r[2] = r[3] = 0; i < l; i++) {
      r[0] += u[i] - v[i] + w[i];
      r[1] +=        v[i] + w[i];
      r[2] += u[i] + v[i];
      r[3] += u[i]        - w[i];
      u[i] = r[0] & mask; r[0] >>= shift;
      v[i] = r[1] & mask; r[1] >>= shift;
      w[i] = r[2] & mask; r[2] >>= shift;
      x[i] = r[3] & mask; r[3] >>= shift;
    }
  }
}


                   /* +----------------------------------+
                      |  Soustraction modulo base^p - 1  |
                      +----------------------------------+ */

/* Entres
     1 <= n,p <= base
     a de longueur p
     b de longueur n

   Sortie
     a <- b - a mod (base^p - 1) de longueur p

   Remarque
     le rsultat n'est pas normalis
*/
void dec_bp(ndouble *a, ndouble *b, int p, int n) {
  int i;
  zdouble z;

  /* retranche le dernier bloc */
  if (n%p) {
    i = n - (n%p); b += i; n -= i;
    z =  neg(a, a, p);
    z += inc(a, b, p, n);
  } else {
    i = n-p; b += i;
    z = sub(a, b, a, p, p);
  }

  /* additionne les blocs prcdents */
  for (; i; i-=p) {b -= p; z += inc(a,b,p,p);}

  /* rincopore la retenue */
  while (z > 0)   z =  inc(a, (ndouble *)(&z), p, 1);
  for (z = -z; z; z = -dec(a, (ndouble *)(&z), p, 1));

}


     /* +---------------------------------------------------------+
        |  Rduction modulo ppcm(base^p - 1, .., base^(p-q) - 1)  |
        +---------------------------------------------------------+ */

/* Entres
    1 <= n,p <= base
    q >= 0
    p >  q, tq p+q est divisible par q!
    b de longueur n

  Sorties
    a <- b mod ppcm(base^p - 1, .., base^(p+q) - 1) de longueur (q+1)*(p+q/2)
    b <- indtermin

  Remarque
    ppcm = produit/(base - 1)..(base^q - 1)

*/
void mod_bq(ndouble *a, ndouble *b, int p, int q, int n) {

  /* cas de base : a <- b mod base^p - 1 */
  if (q == 0) {
    if (n) mod_bp(a,b,p,n);
    else clearmem(a,p);
  }

  /* cas rcursif */
  else {
    ndouble *buff = getmem((n+q+1 < p+1) ? p+1 : n+q+1); 
    ndouble r;
    int i,l;

    /* b <- b div (B^p - 1)/(B^q - 1), buff[q..p] <- reste */
    if (n > p-q) {
      /* buff <- b*(B^q - 1) */
      movemem(buff+q, b, n);
      clearmem(buff,q);
      dec(buff, b, n+q, n);
      n += q+1;
      buff[n-1] = 0;

      /* divise par B^p - 1, b <- quotient, buff[0..p-1] <- reste */
      for (i = (n%p) ? n - (n%p) : n - p; i >= p; i -= p) {
        r = inc(buff+i-p, buff+i, p, (n-i > p) ? p : n-i);
        while(r) {inc(buff+i,&r,n-i,1); r = inc(buff+i-p,&r,p,1);}
      }
      for (i = 0; (i < p) && (buff[i] == mask); i++);
      if (i >= p) {clearmem(buff,p); inc(buff+p,&un,n-p,1);}
      for (n -= p; (n > 0) && (buff[n+p-1] == 0); n--);
      movemem(b,buff+p,n);

      /* divise le reste par base^q - 1 */
      buff[p] = 0;
      for (i = p - q; i >= q; i -= q) {
        r = inc(buff+i-q, buff+i, q, q);
        while(r) {inc(buff+i,&r,p+1-i,1); r = inc(buff+i-q,&r,q,1);}
      }
      for (i = 0; (i < q) && (buff[i] == mask); i++);
      if (i >= q) {inc(buff+q,&un,p+1-q,1);}

    }
    else {
      if (n) movemem(buff+q, b, n);
      clearmem(buff+q+n, p-q+1-n);
      n = 0;
    }

    /* a <- b mod ppcm(B^(p+1) - 1, .., B^(p+q) - 1) */
    mod_bq(a, b, p+1, q-1, n);

    /* a <- a*(B^p - 1)/(B^q - 1) */
    for (l = ((2*p+q+1)*q)/2; (l > 0) && (a[l-1] == 0); l--);
    if (l) {
      movemem(a+p,a,l);
      clearmem(a,p);
      dec(a,a+p,l+p,l);
      l += p+1;
      a[l-1] = 0;
      for (i = (l%q) ? l - (l%q) : l - q; i >= q; i -= q) {
        r = inc(a+i-q, a+i, q, (l-i > q) ? q : l-i);
        while(r) {inc(a+i,&r,l-i,1); r = inc(a+i-q,&r,q,1);}
      }
      for (i = 0; (i < q) && (a[i] == mask); i++);
      if (i >= q) {inc(a+q,&un,l-q,1);}
      movemem(a,a+q,l-q);
      for (l -= q; (l>0) && (a[l-1] == 0); l--);
    }

    /* a <- a + buff[q..p] */
    clearmem(a+l,((q+1)*(2*p+q))/2-l);
    inc(a, buff+q, ((q+1)*(2*p+q))/2, p-q+1);

    freemem(buff);
  }

}



               /* +--------------------------------+
                  |  Recomposition modulo 4^p - 1  |
                  +--------------------------------+ */

/* Entres
     1 <= p <= base
     a de longueur ceil((p+1)/shift) + ceil(p/shift),
     a = [u,v] avec u = x mod (2^p + 1), v = x mod (2^p - 1)

   Sorties
     a <- x mod (4^p - 1)

   Remarque
     Le rsultat est cod sur ceil(2p/shift) chiffres, il n'est pas normalis.
*/
void mjoin(ndouble *a, int p) {

  /* si p bits = nb entier de chiffres -> addition/soustraction */
  if ((p%shift) == 0) {
    int i, l = p/shift;
    ndouble *b = a+l+1, r, ret = a[l];
    zdouble z;

    /* a <- a+b, b <- b - a */
    for (i=r=z=0; i < l; i++) {
      r += a[i] + b[i]; z += b[i] - a[i];
      a[i] = r & mask;  b[i-1] = z & mask;
      r >>= shift;      z >>= shift;
    }

    /* propage les retenues */
    r += ret;  z += inc(b-1,&r,l,1) - ret;
    while (z > 0)     {z =  inc(a, (ndouble *)(&z), 2*l,1);}
    z = -z; while (z) {z = -dec(a, (ndouble *)(&z), 2*l,1);}

    /* divise par 2 */
    for (r = a[0] & 1, i=2*l-1; i >= 0; i--) {
      r = (r << shift) + a[i];
      a[i] = r/2;
      r &= 1;
    }
  }

  /* sinon, addition/soustraction + dcalages */
  else {
    int i, l = p/shift, s = shift - (p%shift);
    ndouble *b = a+l+1, r;
    zdouble z;

    /* a <- a+b, b <- b - a */
    for (i=r=z=0; i <= l; i++) {
      r += a[i] + b[i]; z += b[i] - a[i];
      a[i] = r & mask;  b[i] = z & mask;
      r >>= shift;      z >>= shift;
    }

    /* propage les retenues et dcale b de shift-s bits */
    r = (r << s) + (a[l] >> (shift-s));
    z += inc(b,&r,l+1,1);
    z = (z << s) + (b[l] >> (shift-s));
    b[l] &= (1 << (shift-s)) - 1;
    for (i=l, r=0; i>=0; i--) {
      r = (r << shift) + b[i];
      b[i] = (r >> s) & mask;
    }
    r = (a[l] & ((1 << (shift-s)) - 1)) + (r << (shift-s));
    a[l] = r & mask;

    /* rincorpore les bits surnumraires */
    l = (2*p+shift-1)/shift; s = l*shift - 2*p;
    while (z > 0)    {z =   inc(a, (ndouble *)(&z), l,1)  << s;}
    z = -z; while(z) {z = (-dec(a, (ndouble *)(&z), l,1)) << s;}

    /* divise par 2 */
    for (i=l-1, r = a[i] + ((a[0] & 1) << (shift-s)); i; i--) {
      a[i] = r/2;
      r = ((r & 1) << shift) + a[i-1];
    }
    a[0] = r/2;
  }
}


                    /* +--------------------------------+
                       |  Recomposition modulo 8^p + 1  |
                       +--------------------------------+ */

/* Entres
     p >= shift
     a = [s,u,v,w] de longueur 4*(2*ceil(p/shift)+1) (nombres en compl.  2)
     avec s + u + v + w positif divisible par 3

   Sortie
     c <- ((s+u+v+w)/3)*(1 - 2^p + 4^p) - (u - v*2^p + w*4^p) + mod (8^p + 1)
     de longueur ceil((3*p+1)/shift)
*/

void mcomb3(ndouble *c, ndouble *a, int p) {
  int l = 2*((p+shift-1)/shift) + 1, i, m, n, sh;
  ndouble r, *u = a+l, *v = u+l, *w = v+l, *x;
  zdouble z;

  /* s <- (s + u + v + w)/3 */
  inc(a, u, l, l);
  inc(a, v, l, l);
  inc(a, w, l, l);
  for (i=l-1, r=0; i >= 0; i--) {r = (r << shift) + a[i]; a[i] = r/3; r %= 3;}

  /* u <- s - u,  v <- v - s,  w <- s - w */
  sub(u, a, u, l, l);
  dec(v, a,    l, l);
  sub(w, a, w, l, l);

  /* u <- u + v*2^p */
  m = p/shift; sh = p%shift;
  if (sh) {
    if (u[l-1] & msb) {r = ((ndouble)1) << (shift-sh); dec(v+l-m-1, &r, m+1, 1);}
    for (i=0, x=u+m, r=0; i < l-m; i++) {
      r += x[i] + (v[i] << sh);
      x[i] = r & mask;
      r >>= shift;
    }
    for (; i < l; i++) {
      r += (v[i] << sh);
      x[i] = r & mask;
      r >>= shift;
    }
    if (v[l-1] & msb) x[l] = (r + (mask << sh)) & mask; else x[l] = r;
    n = l+m+1;
  }
  else {
    if (u[l-1] & msb) {dec(v+l-m, &un, m, 1);}
    add(u+m, u+m, v, l-m, l);
    n = l+m;
  }

  /* u <- u + w*4^p */
  m = (2*p)/shift; sh = (2*p)%shift;
  if (sh) {
    if (u[n-1] & msb) {r =  ((ndouble)1) << (shift-sh); dec(w+n-m-1, &r, l-n+m+1, 1);}
    for (i=0, x=u+m, r=0; i < n-m; i++) {
      r += x[i] + (w[i] << sh);
      x[i] = r & mask;
      r >>= shift;
    }
    for (; i < l; i++) {
      r += (w[i] << sh);
      x[i] = r & mask;
      r >>= shift;
    }
    if (w[l-1] & msb) x[l] = (r + (mask << sh)) & mask; else x[l] = r;
    n = l+m+1;
  }
  else {
    if (u[n-1] & msb) {dec(w+n-m, &un, l-n+m, 1);}
    add(u+m, u+m, w, n-m, l);
    n = l+m;
  }

  /* c <- u mod 8^p + 1 */
  m = (3*p+shift)/shift; sh = m*shift - 3*p;

  if (sh < shift) {
    for (i=0, x=u+m, z=0; i < n-m; i++) {
      z += u[i] - (x[i] << sh);
      c[i] = z & mask;
      z >>= shift;
    }
    if (u[n-1] & msb) z -= ((-1) << sh);

    for (; i < m; i++) {
      z += u[i];
      c[i] = z & mask;
      z >>= shift;
    }
  } else {
    c[0] = u[0];
    z = sub(c+1, u+1, u+m, m-1, n-m);
    if (u[n-1] & msb) z += inc(c+n-m+1, &un, 2*m-n-1, 1);
  }

  if (z < 0) { z = (-z) << sh; z = inc(c, (ndouble *)(&z), m, 1);}
  sh = (3*p)%shift;
  z = (z << shift) + c[m-1];
  r = (z >> sh) - 1;
  if (r) {
    c[m-1] = (z - (r << sh)) & mask;
    dec(c, &r, m, 1);
  }

}


      /* +---------------------------------------------------+
         |  Recomposition modulo ppcm(B^p - 1, B^(p+1) - 1)  |
         +---------------------------------------------------+ */

/* Entres
    1 <= p <= base
    a de longueur 2p+1
    a = [u,v] avec u = x mod (B^(p+1) - 1), v = x mod (B^p - 1)

  Sortie
    a <- x mod ppcm(B^p - 1, B^(p+1) - 1)

  Remarques
    ppcm = (B^p - 1)*(B^(p+1) - 1)/(B - 1)
    le rsultat est normalis
*/

void mjoin1(ndouble *a, int p) {
  int i;
  ndouble *b = a + p+1, r;

  /* normalise a */
  for (i=0; (i <= p) && (a[i] == mask); i++); if (i > p) clearmem(a, p+1);

  /* b <- b - a mod (B^p - 1), normalis */
  r = a[p] - dec(b, a, p, p);
  while (r) {r = -dec(b, &r, p, 1);}
  for (i=0; (i < p) && (b[i] == mask); i++); if (i >= p) clearmem(b,p);

  /* b <- b/(B - 1) */
  for (i=p-1, r=0; i >= 0; i--) {
    r = (r << shift) + b[i];
    b[i] = r/mask;
    r %= mask;
  }

  /* a <- a + (B^(p+1) - 1)*b */
  dec(a, b, 2*p+1, p);
}


   /* +----------------------------------------------------------------+
      |  Recomposition modulo ppcm(B^p - 1, B^(p+1) - 1, B^(p+2) - 1)  |
      +----------------------------------------------------------------+ */

/* Entres
    4 <= p <= base, p pair
    a de longueur 3p+3
    a = [u,v,w] avec u = x mod (B^(p+2) - 1),
                     v = x mod (B^(p+1) - 1),
                     w = x mod (B^p     - 1)

  Sortie
    a <- x mod ppcm(B^p - 1, B^(p+1) - 1, B^(p+2) - 1)

  Remarques
    ppcm = (B^p - 1)*(B^(p+1) - 1)*(B^(p+2) - 1)/(B - 1)/(B^2 - 1)
    le rsultat est normalis
*/

void mjoin2(ndouble *a, int p) {
  int i,j;
  ndouble *b = a + p+2, *c = b + p+1, r;
  zdouble z,z0,z1,z2,z3;

  /* normalise a */
  for (i=0; (i <= p+1) && (a[i] == mask); i++); if (i > p+1) clearmem(a, p+2);

  /* b <- b - a mod (B^(p+1) - 1), normalis */
  r = a[p+1] - dec(b, a, p+1, p+1);
  while (r) {r = -dec(b, &r, p+1, 1);}
  for (i=0; (i <= p) && (b[i] == mask); i++); if (i > p) clearmem(b,p+1);

  /* b <- b/(B - 1) */
  for (i=p, r=0; i >= 0; i--) {
    r = (r << shift) + b[i];
    b[i] = r/mask;
    r %= mask;
  }

  /* c <- c - a - (B^2 - 1)*b normalis */
  z  = dec(c,   a, p,   p)   + dec(c, a+p, p, 2);
  z += inc(c,   b, p,   p)   + b[p];                  
  z += dec(c+2, b, p-2, p-2) + dec(c, b+p-2, p, 3);
  while (z > 0) z =  inc(c, (ndouble *)(&z), p, 1);
  for (z=-z; z; z = -dec(c, (ndouble *)(&z), p, 1));
  for (i=0; (i < p) && (c[i] == mask); i++); if (i >= p) clearmem(c,p);

  /* c <- c/(B^2 - 1)/(B - 1), r <- reste */
  z1 = c[p-1]; z2 = c[p-2]; z3 = c[p-3];
  c[p-1] = c[p-2] = c[p-3] = 0;

  for (i = p-4; i >= 0; i--) {
    z0 = r = z1; z1 = z2; z2 = z3; z3 = c[i];
    while(z0) {
      z3 -= z0;
      z2 += z0 + (z3 >> shift); z3 &= mask;
      z1 += z0 + (z2 >> shift); z2 &= mask;
      z0  = z1 >> shift;        z1 &= mask;
      r  += z0;
    };
    for (c[i] = r & mask, r >>= shift, j=i+1; r; j++) {
      r   += c[j];
      c[j] = r & mask;
      r  >>= shift;
    }
  }
  z3 += z1;
  z2 += z3 >> shift;
  r   = z1 + (z2 >> shift);
  if (z3 & mask) r++;

      
  /* b <- b + (B^(p+1) - 1)*c + (B^(p+1) - 1)/(B - 1)*r */
  for (i=z=0; i < p; i++) {
    z += b[i] - c[i] + r;
    b[i] = z & mask;
    z >>= shift;
  }
  z += r;
  if (z > 0)    inc(b+p, (ndouble *)(&z), p+1, 1);
  else {z = -z; dec(b+p, (ndouble *)(&z), p+1, 1);}

  /* a <- a + (B^(p+2) - 1)*b */
  dec(a, b, 3*p+3, 2*p+1);

}


     /* +-------------------------------------------------------+
        |  Recomposition modulo ppcm(B^p - 1, .., B^(p+3) - 1)  |
        +-------------------------------------------------------+ */

/* Entres
    9 <= p <= base, p impair et multiple de 3
    a de longueur 4p+6
    a = [u,v,w,x] avec u = X mod (B^(p+3) - 1),
                       v = X mod (B^(p+2) - 1),
                       w = X mod (B^(p+1) - 1),
                       x = X mod (B^p     - 1)

  Sortie
    a <- x mod ppcm(B^p - 1, .., B^(p+3) - 1)

  Remarques
    ppcm = produit/(B - 1).(B^2 - 1).(B^3 - 1)
    le rsultat est normalis
*/

void mjoin3(ndouble *a, int p) {
  int i,j;
  ndouble *b = a + p+3, *c = b + p+2, *d = c + p+1, r;
  zdouble z,z0,z1,z2,z3;

  /* normalise a */
  for (i=0; (i <= p+2) && (a[i] == mask); i++); if (i > p+2) clearmem(a, p+3);

  /* b <- b - a mod (B^(p+2) - 1), normalis */
  r = a[p+2] - dec(b, a, p+2, p+2);
  while (r) {r = -dec(b, &r, p+2, 1);}
  for (i=0; (i <= p+1) && (b[i] == mask); i++); if (i > p+1) clearmem(b,p+2);

  /* b <- b/(B - 1) */
  for (i=p+1, r=0; i >= 0; i--) {
    r = (r << shift) + b[i];
    b[i] = r/mask;
    r %= mask;
  }

  /* c <- c - a - (B^2 - 1)*b normalis */
  z  = dec(c,   a, p+1, p+1) + dec(c, a+p+1, p+1, 2);
  z += inc(c,   b, p+1, p+1) + b[p+1];                  
  z += dec(c+2, b, p-1, p-1) + dec(c, b+p-1, p+1, 3);
  while (z > 0) z =  inc(c, (ndouble *)(&z), p+1, 1);
  for (z=-z; z; z = -dec(c, (ndouble *)(&z), p+1, 1));
  for (i=0; (i <= p) && (c[i] == mask); i++); if (i > p) clearmem(c,p+1);

  /* c <- c/(B^2 - 1). Rmq: c < B^p */
  z0 = c[p]; z1 = c[p-1];
  c[p] = c[p-1] = 0;
  for (i = p-2; i >= 0; i--) {
    z = r = z0; z0 = z1; z1 = c[i];
    while(z) {
      z1 += z;
      z0 += z1 >> shift; z1 &= mask;
      z   = z0 >> shift; z0 &= mask;
      r  += z;
    };
    for (c[i] = r & mask, r >>= shift, j=i+1; r; j++) {
      r   += c[j];
      c[j] = r & mask;
      r  >>= shift;
    }
  }
  if (z0) inc(c, &un, p+1, 1);

  /* d <- d - a - (B^3-1)*b - (B^3-1)*(B+1)*c normalis */
  z  = dec(d,   a, p,   p)   + dec(d, a+p,   p, 3);
  z += inc(d,   b, p,   p)   + inc(d, b+p,   p, 2);
  z += dec(d+3, b, p-3, p-3) + dec(d, b+p-3, p, 5);
  z += inc(d,   c, p,   p);
  z += inc(d+1, c, p-1, p-1) + c[p-1];
  z += dec(d+3, c, p-3, p-3) + dec(d, c+p-3, p, 3);
  z += dec(d+4, c, p-4, p-4) + dec(d, c+p-4, p, 4);
  while (z > 0) z =  inc(d, (ndouble *)(&z), p, 1);
  for (z=-z; z; z = -dec(d, (ndouble *)(&z), p, 1));
  for (i=0; (i < p) && (d[i] == mask); i++); if (i >= p) clearmem(d,p);

  /* d <- d/(B^3 - 1). Rmq: d < B^(p-2) */
  z1 = d[p-1]; z2 = d[p-2]; z3 = d[p-3];
  d[p-1] = d[p-2] = d[p-3] = 0;
  for (i = p-4; i >= 0; i--) {
    z = r = z1; z1 = z2; z2 = z3; z3 = d[i];
    while(z) {
      z3 += z;
      z2 += z3 >> shift; z3 &= mask;
      z1 += z2 >> shift; z2 &= mask;
      z   = z1 >> shift; z1 &= mask;
      r  += z;
    };
    for (d[i] = r & mask, r >>= shift, j=i+1; r; j++) {
      r   += d[j];
      d[j] = r & mask;
      r  >>= shift;
    }
  }
  if (z3) inc(d, &un, p, 1);

  /* d <- d/(B^2 - 1), z0,z1 <- reste */
  z0 = d[p-2]; z1 = d[p-3];
  d[p-2] = d[p-3] = 0;
  for (i = p-4; i >= 0; i--) {
    z = r = z0; z0 = z1; z1 = d[i];
    while(z) {
      z1 += z;
      z0 += z1 >> shift; z1 &= mask;
      z   = z0 >> shift; z0 &= mask;
      r  += z;
    };
    for (d[i] = r & mask, r >>= shift, j=i+1; r; j++) {
      r   += d[j];
      d[j] = r & mask;
      r  >>= shift;
    }
  }

  /* c <- c + (B^(p+1) - 1)*d + (B^(p+1) - 1)/(B^2 - 1)*(z1 + B*z0) */
  dec(c, d, 2*p+1, p-1);
  for (i=r=0; i < p; i+=2) {
    r += z1 + c[i];   c[i]   = r & mask; r >>= shift;
    r += z0 + c[i+1]; c[i+1] = r & mask; r >>= shift;
  }

  /* c <- c/(B-1), r <- reste */
  for (r = 0, i=2*p; i >= 0; i--) {
    r = (r << shift) + c[i];
    c[i] = r/mask;
    r %= mask;
  }

  /* b <- b + (B^(p+2) - 1)*c + (B^(p+2) - 1)/(B - 1)*r */
  dec(b, c, 3*p+3, 2*p+1);
  for (i=z=0; i <= p+1; i++) {
    z += b[i] + r;
    b[i] = z & mask;
    z >>= shift;
  }
  inc(b+p+2, (ndouble *)(&z), 2*p+1, 1);

  /* a <- a + (B^(p+3) - 1)*b */
  dec(a, b, 4*p+6, 3*p+3);

}

            /* +---------------------------------------+
               |  Compare deux rsidus modulo B^p - 1  |
               +---------------------------------------+ */

int mcompare(ndouble *a, ndouble *b, int p) {
  int i;

  /* compare les rsidus non normaliss */
  if (compare(a,b,p) == 0) return(0);
  
  /* si diffrents, vrifie que chacun est gal = 0 ou B^p - 1 */
  if (a[0]) {
    for (i=0; (i < p) && (a[i] == mask); i++);
  } else {
    for (i=0; (i < p) && (a[i] == 0); i++);
  }
  if (i < p) return(1);

  if (b[0]) {
    for (i=0; (i < p) && (b[i] == mask); i++);
  } else {
    for (i=0; (i < p) && (b[i] == 0); i++);
  }
  if (i < p) return(1);

  return(0);
}
    
