/* +------------------------------------------------------------------------+
   |                                                                        |
   |                                Carr modulaire                         |
   |                                                                        |
   +------------------------------------------------------------------------+ */

#include "defs.h"


                  /* +---------------------------+
                     |  Carr modulo base^p - 1  |
                     +---------------------------+ */

/* Entres
    1 <= n <= base
    p >= 1
    a de longueur n

   Sortie
    c <- a^2 mod (base^p - 1) de longueur p + shift

   Remarque
    le rsultat tient sur p chiffres, mais n'est pas normalis
*/

void sqr_bp(ndouble *c, ndouble *a, int p, int n, int n0) {
  ndouble *buff = getmem(2*p+2 + (7*p/6+5)), *x, *y;
  zdouble ra[4];
  int i,k,l,m,q;

  /* rduit l'argument modulo base^p - 1 */
  mod_bp(buff+2*p+2, a, p, n); a = buff + 2*p+2; a[p] = 0;

  /* dcoupe en deux tant que le nb de bits est pair et que l'oprande
     est assez grand. Accumule les carrs des rsidus modulo 2^q+1 dans c */
  for (q = p*shift; ((q/shift) > n0) && ((q & 1) == 0); ) {

    /* dcoupage */
    q >>= 1; l = (q+shift-1)/shift;
    msplit(a,q);

    /* lvation au carr des rsidus modulo 2^q + 1 : normalement ils tiennent
       sur l chiffres sauf si q est un multiple de shift et l'oprande
       est gal  2^q mod (2^q + 1). Dans ce cas, on calcule le carr
        la main. */
    if (((q%shift) == 0) && (a[2*l])) {
      clearmem(buff+1,2*l);
      buff[0] = 1;
      mod_2p(c, buff, q, -1);
    }

    /* q non divisible par 3 ou trop petit -> carr ordinaire et rduction */
    else if ((q%3) || (q < 3*shift)) {
      sqr_r(buff, a+l, l, n0);
      buff[2*l] = 0;
      mod_2p(c, buff, q, -1);
    }

    /* q = 3k -> calcul modulo (2^k + 1) et (4^k - 2^k + 1) puis combinaison */
    else {
      k = q/3; m = (k+shift-1)/shift;
      msplit3(a+l, ra, k);
      for (i=0, x=buff, y=a+l; i < 4; i++, x += 2*m+1, y += m) {
	sqr_r(x, y, m, n0);
	x[2*m] = (ra[i]*ra[i]) & mask;
	if (ra[i] < 0) {dec(x+m, y, m+1, m); dec(x+m, y, m+1, m);}
	if (ra[i] > 0) {inc(x+m, y, m+1, m); inc(x+m, y, m+1, m);}
      }
      mcomb3(c, buff, k);
    }

    a[l] = 0;
    c += q/shift + 1;
  }

  /* calcule le dernier carr modulo 2^q - 1. On pourrait gagner un
     peu lorsque q = 3k en calculant les carrs modulo 2^k - 1 et
     4^k + 2^k + 1, mais ici soit on a divis p par au moins 16, soit
     q est trop petit, dans les deux cas le gain est minime et ne vaut
     pas la complication du code. */
  l  = (q+shift-1)/shift;
  sqr_r(buff, a, l, n0); buff[2*l] = 0;
  mod_2p(c, buff, q, 1);

  /* recompose le produit */
  for (; q < p*shift; q <<= 1) {c -= q/shift + 1; mjoin(c,q);}

  freemem(buff);

}

          /* +------------------------------------------+
             |  Carr mod ppcm(base^p-1, base^(p+1)-1)  |
             +------------------------------------------+ */

/* Entres
    n >= 1
    p >= 1
    a de longueur n

   Sortie
    c <- a^2 mod ppcm(base^p - 1, base^(p+1) - 1) de longueur 2p+1 + shift

   Remarques
    ppcm = (base^p - 1)*(base^(p+1) - 1)/(base - 1)
    le rsultat est normalis

*/
void sqr_p1(ndouble *c, ndouble *a, int p, int n, int n0) {

  sqr_bp(c,     a, p+1, n, n0);
  sqr_bp(c+p+1, a, p,   n, n0);
  mjoin1(c,p);

}

   /* +--------------------------------------------------------+
      |  Carr mod ppcm(base^p-1, base^(p+1)-1, base^(p+2)-1)  |
      +--------------------------------------------------------+ */

/* Entres
    n >= 1
    p >= 4 pair
    a de longueur n

   Sortie
    c <- a^2 mod ppcm(base^p - 1, base^(p+1) - 1, base^(p+2) - 1)
                de longueur 3p+3 + shift

  Remarques
    ppcm = (B^p - 1)*(B^(p+1) - 1)*(B^(p+2) - 1)/(B - 1)/(B^2 - 1)
    le rsultat est normalis

*/
void sqr_p2(ndouble *c, ndouble *a, int p, int n, int n0) {

  sqr_bp(c,       a, p+2, n, n0);
  sqr_bp(c+p+2,   a, p+1, n, n0);
  sqr_bp(c+2*p+3, a, p,   n, n0);
  mjoin2(c,p);

}

        /* +----------------------------------------------+
           |  Carr mod ppcm(base^p-1, .., base^(p+3)-1)  |
           +----------------------------------------------+ */

/* Entres
    n >= 1
    p >= 9 impair multiple de 3
    a de longueur n

   Sortie
    c <- a^2 mod ppcm(base^p - 1, .., base^(p+3) - 1)
                de longueur 4p+6 + shift

  Remarques
    ppcm = produit/(B - 1).(B^2 - 1).(B^3 - 1)
    le rsultat est normalis
*/
void sqr_p3(ndouble *c, ndouble *a, int p, int n, int n0) {

  sqr_bp(c,       a, p+3, n, n0);
  sqr_bp(c+p+3,   a, p+2, n, n0);
  sqr_bp(c+2*p+5, a, p+1, n, n0);
  sqr_bp(c+3*p+6, a, p,   n, n0);
  mjoin3(c,p);

}

                      /* +------------------+
                         |  Carr rcursif  |
                         +------------------+ */

/* Entres
    n >= 1
    a de longueur n

  Sortie
    c = a^2 de longueur 2*n
*/

#ifdef use_mul_r
void sqr_r(ndouble *c, ndouble *a, int n, int n0) {

  /* petits oprandes -> carr usuel */
  if (n <= 8*n0) karasqr(c,a,n,n0);

  /* sinon, calcul modulaire en 4 points */
  else {
    int p;
    ndouble *buff;

    /* on veut p = 3 mod 6, p >= 9, 4*p >= 2*n */
    p = (2*n+3)/4; if (p < 9) p = 9; if ((p+3)%6) p += 6 - ((p+3)%6);
    buff = getmem(4*p+6+shift);

    /* carr modulaire */
    sqr_p3(buff, a, p, n, n0);    
    movemem(c, buff, 2*n);

    freemem(buff);
  }
    
}
#endif


                           /* +---------+
                              |  tests  |
                              +---------+ */

/* carr modulo B^p - 1 */
static void size0(int n, int *sa, int *sb, int *sc, int *sd, int *p) {
  int l;

  *p = n; l = *p;

  *sa = (5*n + 1)/2;
  *sb = -(l + shift);
  *sc = 0;
  *sd = 0;
}

void msqrsz0(int n, int *sa, int *sb, int *sc, int *sd) {
  int p;
  size0(n, sa, sb, sc, sd, &p);
}

void msqr0(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {
  int sa, sb, sc, sd, p;

  size0(n, &sa, &sb, &sc, &sd, &p);
  sqr_bp(b, a, p, sa, n0);
}

int msqrchk0(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {
  ndouble *e, *f;
  int sa, sb, sc, sd, p, res;

  size0(n, &sa, &sb, &sc, &sd, &p);
  e = getmem(2*sa);
  f = getmem(-sb);

  karasqr(e, a, sa, n0);
  mod_bp(f, e, p, 2*sa);  res = mcompare(b, f, p);

  freemem(e); freemem(f);

  return(res);
}

/* carr modulo B^p - 1, B^(p+1) - 1 */
static void size1(int n, int *sa, int *sb, int *sc, int *sd, int *p) {
  int l;

  *p = (n+1)/2; l = 2*(*p)+1;

  *sa = (5*n + 1)/2;
  *sb = -(l + shift);
  *sc = 0;
  *sd = 0;
}

void msqrsz1(int n, int *sa, int *sb, int *sc, int *sd) {
  int p;
  size1(n, sa, sb, sc, sd, &p);
}

void msqr1(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {
  int sa, sb, sc, sd, p;

  size1(n, &sa, &sb, &sc, &sd, &p);
  sqr_p1(b, a, p, sa, n0);
}

int msqrchk1(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {
  ndouble *e, *f, *g;
  int sa, sb, sc, sd, p, res;

  size1(n, &sa, &sb, &sc, &sd, &p);
  e = getmem(2*sa);
  f = getmem(p + 1 + shift);
  g = getmem(p + 1 + shift);

  karasqr(e, a, sa, n0);
  mod_bp(f, e, p,   2*sa); mod_bp(g, b, p,   2*p+1); res  = mcompare(f, g, p);
  mod_bp(f, e, p+1, 2*sa); mod_bp(g, b, p+1, 2*p+1); res |= mcompare(f, g, p+1);

  freemem(e); freemem(f); freemem(g);

  return(res);
}

/* carr modulo B^p - 1, B^(p+1) - 1, B^(p+2) - 1 */
static void size2(int n, int *sa, int *sb, int *sc, int *sd, int *p) {
  int l;

  *p = (n+2)/3; *p += (*p%2); if (*p < 4) *p=4; l = 3*(*p)+3;

  *sa = (5*n + 1)/2;
  *sb = -(l + shift);
  *sc = 0;
  *sd = 0;
}

void msqrsz2(int n, int *sa, int *sb, int *sc, int *sd) {
  int p;
  size2(n, sa, sb, sc, sd, &p);
}

void msqr2(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {
  int sa, sb, sc, sd, p;

  size2(n, &sa, &sb, &sc, &sd, &p);
  sqr_p2(b, a, p, sa, n0);
}

int msqrchk2(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {
  ndouble *e, *f, *g;
  int sa, sb, sc, sd, p, res;

  size2(n, &sa, &sb, &sc, &sd, &p);
  e = getmem(2*sa);
  f = getmem(p + 2 + shift);
  g = getmem(p + 2 + shift);

  karasqr(e, a, sa, n0);
  mod_bp(f, e, p,   2*sa); mod_bp(g, b, p,   3*p+3); res  = mcompare(f, g, p);
  mod_bp(f, e, p+1, 2*sa); mod_bp(g, b, p+1, 3*p+3); res |= mcompare(f, g, p+1);
  mod_bp(f, e, p+2, 2*sa); mod_bp(g, b, p+2, 3*p+3); res |= mcompare(f, g, p+2);

  freemem(e); freemem(f); freemem(g);

  return(res);
}

/* carr modulo B^p - 1, B^(p+1) - 1, B^(p+2) - 1, B^(p+3) - 1 */
static void size3(int n, int *sa, int *sb, int *sc, int *sd, int *p) {
  int l;

  *p = (n+3)/4;   while (((*p%6) != 3) || (*p < 9)) (*p)++; l = 4*(*p)+6;

  *sa = (5*n + 1)/2;
  *sb = -(l + shift);
  *sc = 0;
  *sd = 0;
}

void msqrsz3(int n, int *sa, int *sb, int *sc, int *sd) {
  int p;
  size3(n, sa, sb, sc, sd, &p);
}

void msqr3(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {
  int sa, sb, sc, sd, p;

  size3(n, &sa, &sb, &sc, &sd, &p);
  sqr_p3(b, a, p, sa, n0);
}

int msqrchk3(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {
  ndouble *e, *f, *g;
  int sa, sb, sc, sd, p, res;

  size3(n, &sa, &sb, &sc, &sd, &p);
  e = getmem(2*sa);
  f = getmem(p + 3 + shift);
  g = getmem(p + 3 + shift);

  karasqr(e, a, sa, n0);
  mod_bp(f, e, p,   2*sa); mod_bp(g, b, p,   4*p+6); res  = mcompare(f, g, p);
  mod_bp(f, e, p+1, 2*sa); mod_bp(g, b, p+1, 4*p+6); res |= mcompare(f, g, p+1);
  mod_bp(f, e, p+2, 2*sa); mod_bp(g, b, p+2, 4*p+6); res |= mcompare(f, g, p+2);
  mod_bp(f, e, p+3, 2*sa); mod_bp(g, b, p+3, 4*p+6); res |= mcompare(f, g, p+3);

  freemem(e); freemem(f); freemem(g);

  return(res);
}


/* carr complt en 3 points */
static void size4(int n, int *sa, int *sb, int *sc, int *sd, int *p) {
  int l;

  *p = (2*n+2)/3; *p += (*p%2); if (*p < 4) *p=4; l = 3*(*p)+3;

  *sa = n;
  *sb = -(l + shift);
  *sc = 0;
  *sd = 0;
}

void msqrsz4(int n, int *sa, int *sb, int *sc, int *sd) {
  int p;
  size4(n, sa, sb, sc, sd, &p);
}

void msqr4(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {
  int sa, sb, sc, sd, p;

  size4(n, &sa, &sb, &sc, &sd, &p);
  sqr_p2(b, a, p, sa, n0);
}

/* carr complt en 4 points */
static void size5(int n, int *sa, int *sb, int *sc, int *sd, int *p) {
  int l;

  *p = (2*n+3)/4; while (((*p%6) != 3) || (*p < 9)) (*p)++; l = 4*(*p)+6;

  *sa = n;
  *sb = -(l + shift);
  *sc = 0;
  *sd = 0;
}

void msqrsz5(int n, int *sa, int *sb, int *sc, int *sd) {
  int p;
  size5(n, sa, sb, sc, sd, &p);
}

void msqr5(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {
  int sa, sb, sc, sd, p;

  size5(n, &sa, &sb, &sc, &sd, &p);
  sqr_p3(b, a, p, sa, n0);
}

