/* +------------------------------------------------------------------------+
   |                                                                        |
   |                             Oprations en n^2                          |
   |                                                                        |
   +------------------------------------------------------------------------+ */

#include "defs.h"

/* c <- ab, lg(a) = lg(b) = n */
void muln2(ndouble *c, ndouble *a, ndouble *b, int n) {
  int i,j;
  ndouble r;

  clearmem(c,n);
  for (i=0; i < n; i++,a++,c++) {
    for (j=0,r=0; j < n; j++) {
      r   += a[0]*b[j] + c[j];
      c[j] = r & mask;
      r  >>= shift;
    }
    c[n] = r & mask;
  }
}

/* c <- partie basse(a*b), lg(a) = lg(b) = n */
void muln2l(ndouble *c, ndouble *a, ndouble *b, int n) {
  int i,j;
  ndouble r;

  clearmem(c,n+2);
  for (i=0; i < n; i++,a++,c++) {
    for (j=0,r=0; j < n-i; j++) {
      r   += a[0]*b[j] + c[j];
      c[j] = r & mask;
      r  >>= shift;
    }
    r += c[n-i];
    c[n-i]   = r & mask;
    c[n-i+1] = (c[n-i+1] + (r >> shift)) & mask;
  }
}

/* c <- partie haute(a*b), lg(a) = lg(b) = n */
void muln2h(ndouble *c, ndouble *a, ndouble *b, int n) {
  int i,j;
  ndouble r;

  c[0] = 0; c -= n-1; 
  for (i=0; i < n; i++,a++,c++) {
    for (j=n-i-1,r=0; j < n; j++) {
      r   += a[0]*b[j] + c[j];
      c[j] = r & mask;
      r  >>= shift;
    }
    c[n] = r & mask;
  }
}

/* c <- partie mdianne(a*b), lg(a) = 2n-1, lg(b) = n */
void muln2m(ndouble *c, ndouble *a, ndouble *b, int n) {
  int i,j;
  ndouble r;

  clearmem(c,n+2); a += n-1;
  for (i=0; i<n; i++,b++,a--) {
    for (j=0,r=0; j < n; j++) {
      r   += a[j]*b[0] + c[j];
      c[j] = r & mask;
      r  >>= shift;
    }
    r += c[n];
    c[n] = r & mask;
    c[n+1] = (c[n+1] + (r >> shift)) & mask;
  }
}


/* c <- a^2, lg(a) = n */
void sqrn2(ndouble *c, ndouble *a, int n) {
  int i,j;
  ndouble r;

  clearmem(c,2*n);

  /* doubles produits */
  for (i=1; i < n; i++) {
    for (j=0,r=0; j<i; j++) {
      r  += a[i]*a[j] + c[i+j];
      c[i+j] = r & mask;
      r >>= shift;
    }
    c[2*i] = r & mask;
  }
  for (i=0,r=0; i < 2*n; i++) {
    r   += 2*c[i];
    c[i] = r & mask;
    r  >>= shift;
  }

  /* carrs */
  for (i=0,r=0; i < n; i++) {
    r += a[i]*a[i] + c[2*i];
    c[2*i] = r & mask;
    r = (r >> shift) + c[2*i+1];
    c[2*i+1] = r & mask;
    r >>= shift;
  }
}

/* c <- low(a^2), lg(a) = n */
void sqrn2l(ndouble *c, ndouble *a, int n) {
  int i,j;
  ndouble r;

  clearmem(c,n+2);

  /* doubles produits */
  for (i=0; i < n; i++) {
    for (j=i+1,r=0; j < n-i; j++) {
      r  += a[i]*a[j] + c[i+j];
      c[i+j] = r & mask;
      r >>= shift;
    }
    r += c[n]; c[n] = r & mask; c[n+1] = (c[n+1] + (r >> shift)) & mask;
  }
  for (i=0,r=0; i < n+2; i++) {
    r   += 2*c[i];
    c[i] = r & mask;
    r  >>= shift;
  }

  /* carrs */
  for (i=0,r=0; 2*i < n; i++) {
    r += a[i]*a[i] + c[2*i];
    c[2*i] = r & mask;
    r = (r >> shift) + c[2*i+1];
    c[2*i+1] = r & mask;
    r >>= shift;
  }
  for (j=2*i; j < n+2; j++) {
    r += c[j];
    c[j] = r & mask;
    r >>= shift;
  }
}

