/* +------------------------------------------------------------------------+
   |                                                                        |
   |                             Produit mdian                             |
   |                                                                        |
   +------------------------------------------------------------------------+ */

#include "defs.h"

/* c <- mdian(a*b), calcul avec gestion des retenues */
void rmp0(ndouble *c, ndouble *a, ndouble *b, int n, int n0) {

  if (n <= n0) muln2m(c,a,b,n);
  else {
    int     p=n/2, q=n-p;            /* demi-longueurs     */
    ndouble *buff = getmem(n+2*q+6); /* mmoire auxiliaire */
    ndouble *beta = buff+q;
    ndouble *alpha= beta +q+2;
    ndouble *gamma= alpha+q+2;
    ndouble *b00;                    /* b0 ou b0-1         */
    ndouble r0,r1,r2,r;              /* retenues           */
    zdouble s;
     int  i;

    /* dcoupage :
                    <----- 2q-1 ------>  <---- 2p-1 --->
                    +-----------------+--+-------------+
            a:      |      a0         |  |     a2      |
                    +-----------------+--+-------------+
                             +-----------------+
            a10:    <-- q -->|<---- 2q-1 ----->|
                             +-----------------+
                             +-------------+
            a11:    <-- q -->|<-- 2p-1 --->|
                             +-------------+

                    <-- p -->     <-- p -->
                    +--------+----+--------+
            b:      |   b0   | b1 |   b2   |
                    +----- --+----+--------+
                    
                    <------ q+2 ---->
                    +---------------+--------+
            c:      |     alpha     |        |
                    +-------------+-+--------+
                    |             |  gamma   |
                    +-------------+----------+
                                   <-- p+2 -->
    */

    /* rmq : on pourrait calculer a0+a10 et a11+a2 ensemble puisqu'ils
       ont q-1 chiffres en commun, mais s'il y a une retenue qui entre
       dans le bloc commun lors du calcul de a0+a10 a va pas tre coton
       de grer les consquences pour a11+a2. Mieux vaut faire simple
       et sr.                                                        */

     /* rmq : on pourrait calculer alpha et beta dans c au lieu de leur
        rserver de la place dans la mmoire auxiliaire, mais pour une
        raison que je ne m'explique pas, cela ralentit le programme. */

    /* 1. alpha <- rmp(a0,b12) + rmp(a10,b12) */

    /* buff <- a0 + a10 et cumule les retenues */
    for (i=r=r0=r1=0; i < 2*q-1; i++) {
      r += a[i] + a[i+q]; buff[i] = r & mask; r >>= shift;
      if (r) {
        if (i<q-1) r0 += b[n-2-i];
        else       r1 += b[n+q-2-i];
      }
    }

    /* rmp et correction */
    rmp0(alpha, buff, b+p, q, n0);
    dec(alpha,   &r0, q+2, 1);
    inc(alpha+q, &r1,   2, 1);

    /* 2. gamma <- rmp(a11,b0) + rmp(a2,b0) */

    /* buff <- a1 + a12 et cumule les retenues */
    for (i=r=r0=r1=0; i < 2*p-1; i++) {
      r += a[i+q] + a[i+2*q]; buff[i] = r & mask; r >>= shift;
      if (r) {
        if (i<p-1) r0 += b[p-2-i];
        else       r1 += b[2*p-2-i];
      }
    }

    /* rmp et correction */
    rmp0(gamma, buff, b, p, n0);
    dec(gamma,   &r0, p+2, 1);
    inc(gamma+p, &r1,   2, 1);

    /* 3. beta <- rmp(a10,b12) - rmp(a10,b00) */
    /* b00 = b0 prolong par un zro  gauche si n est impair, b0 sinon */

    /* buff <- b12 - b00 et cumule les retenues */
    if (p<q) {b00 = b-1; buff[0] = b[p]; i=1;} else {b00 = b; i=0;}
    for (s=r0=r1=0; i < q; i++) {
      s += b[i+p] - b00[i]; buff[i] = s & mask; s >>= shift;
      if ((s) && (i < q-1)) {
        r0 += a[2*q-2-i];
        r1 += a[3*q-2-i];
      }
    }

    /* rmp et correction */
    rmp0(beta, a+q, buff, q, n0);
    if (s) dec(beta+1, a+q, q+1,q);
    inc(beta,   &r0, q+2, 1);
    dec(beta+q, &r1,   2, 1);

    /* 4. c[0..q+1] <- alpha - beta, sauve les deux derniers chiffres */
    /*    qui vont tre crass par gamma + low(beta)                 */
    sub(c, alpha, beta, q+2, q+2);
    r2 = c[q] + (c[q+1] << shift);

    /* 5. c[q..n+1] <- gamma + low(beta) */
    if (p < q) {
      s = beta[q-1] + (beta[q] << shift) - a[3*q-2]*b[p];
      for (i=1; i<q; i++) s -= a[3*q-2-i]*(b[i+p] - b00[i]);
      beta[q-1] = s & mask;
      beta[q]   = (s >> shift) & mask;
    }
    add(c+q, beta, gamma, p+2, p+2);

    /* 6. rincorpore les chiffres de alpha-beta crass */
    inc(c+q, &r2, p+2, 1);

    /* 7. termin */
    freemem(buff);
  }
}


/* produit mdian par blocs */
void bloc_rmp(ndouble *c, ndouble *a, ndouble *b, int n, int n0) {

  if (n<=1) bloc_muln2(c,a,b,n,n0);
  else {
    int p=n/2, q=n-p;
    ndouble *buff = getmem(3*q*n0);

    bloc_add(buff,      a,         a+p*n0,   n+q-1,n+q-1,n0);
    bloc_rmp(c,         buff,      b+q*n0,   p,n0);          /* u'   + v'    */
    bloc_rmp(c+2*p*n0,  buff+p*n0, b,        q,n0);          /* u:w  + w:v   */
    if (p<q) {
      movemem(buff, b, n0);
      bloc_sub(buff+n0, b+n0,      b+q*n0,   p,p,n0);
    } else {
      bloc_sub(buff,    b,         b+q*n0,   p,p,n0);
    }      
    bloc_rmp(buff+q*n0, a+p*n0,    buff,     q,n0);          /* w:v  - w':u' */
    bloc_inc(c,         buff+q*n0,           p,p,2*n0);      /* v    + v'    */
    bloc_dec(c+2*p*n0,  buff+q*n0,           q,q,2*n0);      /* u:w  + w':u' */

    freemem(buff);
  }
}

void rmp1(ndouble *c, ndouble *a, ndouble *b, int n, int n0) {
  int n1,p;
  ndouble *a1,*b1,*c1;

  /* dcompose en blocs de sorte que le nb de blocs soit proche   */
  /* d'une puissance de 2.                                        */
  for (n1=n; n1>=n0; n1>>=1); n1++; p = (n+n1-1)/n1;

  a1 = getmem(2*p*(n1+1)+1);  expand(a1,a,2*n-1,2*p-1,n1);
  b1 = getmem(p*(n1+1));      expand(b1,b,n,p,n1);
  c1 = getmem(p*2*(n1+1));

  /* calcule le produit mdian par blocs */
  bloc_rmp(c1, a1, b1, p, n1+1);

  /* frontire infrieure */
  if (n1>1) {
    int lg = n-3 - (p-1)*n1; /* largeur troncature           */
    ndouble t[3];              /* somme des produits  retirer */
    int i,j,k,l;
    ndouble r;
    
    clearmem(t,3); 
    for (i=0, k=(p-1)*n1, j=k+lg; i < n; i++,j--) {

      /* high(a[j]b[i]) et a[j+1]b[i]  retrancher */
      if (j+1 >= k) {
        r = a[j+1]*b[i] + t[0];
        if (j >= k) r += ((a[j]*b[i]) >> shift);
        t[0] = r & mask; r = (r >> shift) + t[1];
        t[1] = r & mask; r = (r >> shift) + t[2];
        t[2] = r & mask;
      }

      /* a[j+2]b[i] .. a[k-1]b[i]  ajouter */
      else if (j+2 < k) {
        for (l=j+2, r=0; l < k; l++) {
          r += a[l]*b[i] + c1[lg+l-j];
          c1[lg+l-j] = r & mask;
          r >>= shift;
        }
        for (l=lg+k-j; (r) && (l < 2*n1+2); l++) {
          r += c1[l];
          c1[l] = r & mask;
          r >>= shift;
        }
      }

      /* changement de bloc */
      if ((i+1)%n1 == 0) k -= n1;
    }

    if (lg+1 >= 0) dec(c1+lg+1, t,      2*n1-lg+1, 3);
    else           dec(c1,      t-lg-1, 2*n1+2,    4+lg);

  }

  /* frontire suprieure */
  if (n1>1) {
    ndouble *c2 = c1 + 2*p + 2*n - 3; /* limite de troncature */
    int i,j,k,l;
    ndouble r,t;

    t = 0;
    for(i=0, k=(2*p-1)*n1, j=2*n-2; i < n; i++,j--) {

      /* a[k]b[i] .. a[j]b[i]  ajouter */
      if (j>=k) {
        for (l=k, r=0; l <= j; l++) {
          r += a[l]*b[i] + c2[l-j-1];
          c2[l-j-1] = r & mask;
          r >>= shift;
        }
        r += c2[0]; c2[0] = r & mask; r >>= shift;
        r += c2[1]; c2[1] = r & mask;
      }

      /* a[j+1]b[i] et low(a[j+2]b[i])  retrancher */
      if ((i>0) && (j+1 < k)) {
        t += a[j+1]*b[i];
        if ((i>1) && (j+2 < k)) t += (a[j+2]*b[i]) << shift;
      }

      /* changement de bloc */
      if ((i+1)%n1 == 0) k -= n1;
    }
    t = c2[0] + (c2[1] << shift) - t;
    c2[0] = t & mask;
    c2[1] = (t >> shift) & mask;
  }

  /* compacte le rsultat */
  { 
    int k = p*n1 - n + 1;
    compact(c+k, c1+2*(n1+1), p-1, 2*n1+2, n1, n+2-k);
    clearmem(c,k);
    if (n1+2+k <= n+2) inc(c,c1+n1-k,n+2, n1+2+k);
    else               inc(c,c1+n1-k,n+2, n+2);
  }

  freemem(a1); freemem(b1); freemem(c1);

}

/* tests */
void mpsz(int n, int *sa, int *sb, int *sc, int *sd) {
  *sa = 2*n-1;
  *sb = n;
  *sc = -(n+2);
  *sd = 0;
}

void mp0(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {
  rmp0(c, a, b, n, n0);
}

void mp1(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {
  rmp1(c, a, b, n, n0);
}

int mpchk(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {
  ndouble *e;
  int res;

  e = getmem(n+2);
  muln2m(e,a,b,n);
  res = compare(c,e,n+2);
  freemem(e);

  return(res);
}
